/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.description;

import java.io.InputStream;
import java.io.StringWriter;
import java.util.Collection;
import java.util.Map;

public interface Description {

	/**
	 * Gets the class name of the element.
	 *
	 * @return the fully qualified class name of the element, or <code>null</code> if element is not a widget.
	 */
	String getClassName();

	/**
	 * Gets the element friendly name.
	 *
	 * @return the element friendly name.
	 */
	String getFriendlyName();

	/**
	 * Gets the element description.
	 *
	 * @return the element description.
	 */
	String getDescription();

	/**
	 * Gets the element 16x16 icon.
	 *
	 * @return the element 16x16 icon or <code>null</code> if none.
	 */
	InputStream getIcon16();

	/**
	 * Gets the element 24x24 icon.
	 *
	 * @return the element 24x24 icon or <code>null</code> if none.
	 */
	InputStream getIcon24();

	/**
	 * Gets whether the element is a root one or not.
	 *
	 * @return <code>true</code> if the element is a root one, <code>false</code> otherwise.
	 */
	boolean isRoot();

	/**
	 * Gets whether the attribute is required or not.
	 *
	 * @param attributeName
	 *            the attribute name.
	 * @return <code>true</code> if the attribute is required, <code>false</code> otherwise.
	 */
	boolean isRequiredAttribute(String attributeName);

	/**
	 * Gets the type of an attribute of the element.
	 *
	 * @param attributeName
	 *            the attribute name.
	 * @return the type of the attribute.
	 */
	Type getAttributeType(String attributeName);

	/**
	 * Generates the element initialization.
	 * <p>
	 * Writes the Java code necessary to initialize the element (constructor & setters).
	 *
	 * @param variableName
	 *            the receiver name.
	 * @param attributes
	 *            the attributes name and values.
	 * @param writer
	 *            the writer to write in.
	 */
	void generateInitialization(String variableName, Map<String, String> attributes, StringWriter writer);

	/**
	 * Generates the element initialization.
	 * <p>
	 * Creates an instance the element with the given attributes.
	 *
	 * @param classLoader
	 *            the class loader to use.
	 * @param attributes
	 *            the attributes name and values.
	 * @throws ReflectiveOperationException
	 *             if a problem occurs while instantiating the element.
	 */
	Object newInstance(ClassLoader classLoader, Map<String, Object> attributes) throws ReflectiveOperationException;

	/**
	 * Gets whether or not the child elements order must be maintained.
	 *
	 * @return <code>true</code> if the children order must be maintained, <code>false</code> otherwise.
	 */
	boolean hasOrderedElements();

	/**
	 * Gets the default child element of a container.
	 * <p>
	 * The default element must not have attributes.
	 * <p>
	 * Returns <code>null</code> if there is no default element.
	 *
	 * @return the default child element tag.
	 * @see #getElements()
	 */
	String getDefaultElement();

	/**
	 * Gets the child elements of a container.
	 *
	 * @return the child elements tags.
	 */
	Collection<String> getElements();

	/**
	 * Gets the type of an attribute of a child element.
	 *
	 * @param elementName
	 *            the child element name.
	 * @param attributeName
	 *            the child element attribute name.
	 * @return the type of the attribute.
	 */
	Type getElementAttributeType(String elementName, String attributeName);

	/**
	 * Gets the type of a child element.
	 *
	 * @param elementName
	 *            the child element name.
	 * @return the type of the element.
	 */
	Type getElementType(String elementName);

	/**
	 * Gets the maximum cardinality of a child element.
	 * <p>
	 * If the maximum cardinality is unbounded, returns {@link Integer#MAX_VALUE}.
	 *
	 * @param elementName
	 *            the child element name.
	 * @return the cardinality of the element.
	 */
	int getElementMaxCardinality(String elementName);

	/**
	 * Generates the child element.
	 *
	 * @param variableName
	 *            the receiver name.
	 * @param elementName
	 *            the child element name.
	 * @param attributes
	 *            the child element attributes name and values.
	 * @param childVariableName
	 *            the child variable name.
	 * @param writer
	 *            the writer to write in.
	 */
	void generateChild(String variableName, String elementName, Map<String, String> attributes,
			String childVariableName, StringWriter writer);

	/**
	 * Generates the child element.
	 *
	 * @param classLoader
	 *            the class loader to use.
	 * @param elementName
	 *            the child element name.
	 * @param receiver
	 *            the receiver.
	 * @param attributes
	 *            the child element attributes name and values.
	 * @param child
	 *            the child element.
	 * @throws ReflectiveOperationException
	 *             if a problem occurs while adding the child.
	 */
	void addChild(ClassLoader classLoader, Object receiver, String elementName, Map<String, Object> attributes,
			Object child) throws ReflectiveOperationException;

	/**
	 * Gets the attributes of the element.
	 *
	 * @return the attributes of the element.
	 */
	Collection<String> getAttributes();

	/**
	 * Gets the attributes of the given child element.
	 *
	 * @param elementName
	 *            the child element name.
	 * @return the attributes of the given child element.
	 */
	Collection<String> getElementAttributes(String elementName);

	/**
	 * Gets the element name.
	 *
	 * @return the element name.
	 */
	String getTag();

}
