/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model.test.attribute;

import java.util.HashSet;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.test.AbstractTest;

/**
 * Unit tests for {@link Attribute} class.
 */
public class AttributeTest extends AbstractTest {

	private static final String VALID_NAME = "avalidname";
	private static final String VALID_VALUE = "avalidvalue";
	private static final String VALID_VALUE_INTEGER = "100";
	private static final Integer VALID_VALUE_INTEGER_AS_INTEGER = new Integer(100);
	private static final String VALID_VALUE_FLOAT = "100.1";
	private static final Float VALID_VALUE_FLOAT_AS_FLOAT = new Float(100.1f);
	private static final String VALID_VALUE_DOUBLE = "100.11265456214";
	private static final Double VALID_VALUE_DOUBLE_AS_DOUBLE = new Double(100.11265456214d);

	@Test(expected = IllegalArgumentException.class)
	public void testNullAttributeName() {
		new Attribute(null, VALID_VALUE);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testEmptyAttributeName() {
		new Attribute("", VALID_VALUE);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullAttributeValue() {
		new Attribute(VALID_NAME, null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testEmptyAttributeValue() {
		new Attribute(VALID_NAME, "");
	}

	@Test
	public void testValidAttribute() {
		Attribute attribute = new Attribute(VALID_NAME, VALID_VALUE);
		Assert.assertEquals(attribute.getName(), VALID_NAME);
		Assert.assertEquals(attribute.getValue(), VALID_VALUE);
	}

	@Test
	public void testValidAttributeValueAsInteger() {
		Attribute attribute = new Attribute(VALID_NAME, VALID_VALUE_INTEGER);
		Assert.assertEquals(attribute.getIntegerValue(), VALID_VALUE_INTEGER_AS_INTEGER);
	}

	@Test
	public void testValidAttributeValueAsFloat() {
		Attribute attribute = new Attribute(VALID_NAME, VALID_VALUE_FLOAT);
		Assert.assertEquals(attribute.getFloatValue(), VALID_VALUE_FLOAT_AS_FLOAT);
	}

	@Test
	public void testValidAttributeValueAsDouble() {
		Attribute attribute = new Attribute(VALID_NAME, VALID_VALUE_DOUBLE);
		Assert.assertEquals(attribute.getDoubleValue(), VALID_VALUE_DOUBLE_AS_DOUBLE);
	}

	@Test(expected = NumberFormatException.class)
	public void testInvalidAttributeValueAsInteger() {
		new Attribute(VALID_NAME, VALID_VALUE).getIntegerValue();
	}

	@Test(expected = NumberFormatException.class)
	public void testInvalidAttributeValueAsFloat() {
		new Attribute(VALID_NAME, VALID_VALUE).getFloatValue();
	}

	@Test(expected = NumberFormatException.class)
	public void testInvalidAttributeValueAsDouble() {
		new Attribute(VALID_NAME, VALID_VALUE).getDoubleValue();
	}

	@Test
	public void testAttributeHashSetEquals() {
		Attribute attribute1 = new Attribute(VALID_NAME, VALID_VALUE);
		Attribute attribute2 = new Attribute(VALID_NAME, VALID_VALUE);
		Set<Attribute> attributeSet = new HashSet<Attribute>();
		attributeSet.add(attribute1);

		Assert.assertEquals(attributeSet.iterator().next(), attribute2);
		Assert.assertTrue(attributeSet.contains(attribute2));
	}

	@Test
	public void testAttributeHashSetHashcode() {
		Attribute attribute1 = new Attribute(VALID_NAME, VALID_VALUE);
		Attribute attribute2 = new Attribute(VALID_NAME, VALID_VALUE);
		Set<Attribute> attributeSet = new HashSet<Attribute>();
		attributeSet.add(attribute1);
		attributeSet.add(attribute2);

		Assert.assertEquals(attribute1, attribute2);
		Assert.assertEquals(attributeSet.size(), 1);
	}

	@Test
	public void testAttributeEquals() {
		Attribute attribute1 = new Attribute(VALID_NAME, VALID_VALUE);
		Attribute attribute2 = new Attribute(VALID_NAME, VALID_VALUE);
		Attribute attribute3 = new Attribute(VALID_NAME + "--", VALID_VALUE);
		Attribute attribute4 = new Attribute(VALID_NAME, VALID_VALUE + "--");
		Attribute attribute5 = new Attribute(VALID_NAME.toUpperCase(), VALID_VALUE);

		Assert.assertFalse(attribute1.equals(null));
		Assert.assertFalse(attribute1.equals(new String()));
		Assert.assertFalse(attribute1.equals(attribute3));
		Assert.assertFalse(attribute1.equals(attribute4));
		Assert.assertFalse(attribute3.equals(attribute4));

		Assert.assertTrue(attribute1.equals(attribute2));
		Assert.assertTrue(attribute1.equals(attribute5));

	}

}
