/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;
import com.is2t.giml.xml.parser.exception.XMLParserUnknownAttributeException;

/**
 * The class {@code XMLParserModelFactory} allows to create high-level elements from XML nodes.
 *
 * @see Document
 * @see Node
 * @see Attribute
 */
public class XMLParserModelFactory {

	private XMLParserModelFactory() {
		// utility class
	}

	/**
	 * Creates an attribute from given name and value.
	 *
	 * @param name
	 *            the new attribute name.
	 * @param value
	 *            the new attribute value.
	 * @return the created attribute.
	 * @throws IllegalArgumentException
	 *             if given values are not valid.
	 * @see Attribute#Attribute(String, String)
	 */
	public static Attribute createAttribute(String name, String value) throws IllegalArgumentException {
		return new Attribute(name, value);
	}

	/**
	 * Creates a new attribute from the given XML node describing an attribute.
	 *
	 * @param xmlNode
	 *            the XML node to create an attribute from.
	 * @return the attribute describing the given XML node.
	 * @throws IllegalArgumentException
	 *             if the given XML node does not represent an XML attribute node.
	 * @see Attribute#Attribute(String, String)
	 */
	public static Attribute createAttribute(org.w3c.dom.Node xmlNode) throws IllegalArgumentException {
		if (xmlNode.getNodeType() != org.w3c.dom.Node.ATTRIBUTE_NODE) {
			throw new IllegalArgumentException("Given XML node is not an attribute node");
		}
		return createAttribute(xmlNode.getNodeName(), xmlNode.getNodeValue());
	}

	/**
	 * Create a new Document from the given arguments.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the root tag name of the document.
	 * @param attributes
	 *            the attributes of the root tag of the document
	 * @param children
	 *            child nodes of the document.
	 * @param importedDocuments
	 *            child documents of this document.
	 * @return the newly created document.
	 * @throws IllegalArgumentException
	 *             if given argument are not valid
	 *
	 * @see Document#Document(String, String, Set, List, List)
	 */
	public static Document createDocument(String id, String name, Set<Attribute> attributes, List<Node> children,
			List<Document> importedDocuments) throws IllegalArgumentException {
		return createDocument(id, name, attributes, children, importedDocuments, new ArrayList<Template>());
	}

	/**
	 * Create a new Document from the given arguments.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the root tag name of the document.
	 * @param attributes
	 *            the attributes of the root tag of the document
	 * @param children
	 *            child nodes of the document.
	 * @param importedDocuments
	 *            child documents of this document.
	 * @param templates
	 *            templates declared in the document.
	 * @return the newly created document.
	 * @throws IllegalArgumentException
	 *             if given argument are not valid
	 *
	 * @see Document#Document(String, String, Set, List, List, List)
	 */
	public static Document createDocument(String id, String name, Set<Attribute> attributes, List<Node> children,
			List<Document> importedDocuments, List<Template> templates) throws IllegalArgumentException {
		return new Document(id, name, attributes, children, importedDocuments, templates);
	}

	/**
	 * Creates a new node from given values.
	 *
	 * @param name
	 *            the name of the node to create.
	 * @param attributes
	 *            the set of attributes of the node to create.
	 * @param children
	 *            the list of children of the node to create.
	 * @return the newly created node.
	 * @throws IllegalArgumentException
	 *             if given values are not valid.
	 * @see Node#Node(String, Set, List)
	 */
	public static Node createNode(String name, Set<Attribute> attributes, List<Node> children)
			throws IllegalArgumentException {
		return new Node(name, attributes, children);
	}

	/**
	 * Create a new document from an XML element. The created document does not have resolved imports, thus it does not
	 * have child documents.
	 *
	 * @param xmlElement
	 *            the XML element to create a document from.
	 * @param documentIdAttributeName
	 *            the name of the attribute that describe the document unique ID in the given XML element.
	 * @return the document describing the given XML node
	 * @throws XMLParserUnknownAttributeException
	 *             if the given element does not contains the attribute that describe the document unique ID.
	 * @throws IllegalArgumentException
	 *             if extracted XML element information are not valid to create a document.
	 */
	public static Document createDocument(Element xmlElement, String documentIdAttributeName)
			throws XMLParserUnknownAttributeException, IllegalArgumentException {
		// First, check the XML node is well formed, i.e. it contains at least the attribute describing document ID
		Set<Attribute> attributes = createAttributeSet(xmlElement);
		String documentId = null;
		for (Attribute attribute : attributes) {
			if (documentId == null && attribute.getName().equals(documentIdAttributeName)) {
				documentId = attribute.getValue();
			}
		}
		if (documentId == null) {
			throw new XMLParserUnknownAttributeException(documentIdAttributeName, xmlElement.getTagName(), "(unknown)");
		}

		// Then, create the list of its child nodes
		List<Node> children = new ArrayList<Node>();
		NodeList xmlChildNodes = xmlElement.getChildNodes();
		for (int c = 0; c < xmlChildNodes.getLength(); c++) {
			org.w3c.dom.Node child = xmlChildNodes.item(c);
			if (child instanceof Element) {
				Node createdNode = createNode((Element) child);
				if (createdNode != null) {
					children.add(createdNode);
				}
			}
		}

		// Finally, return the new document. it does not have child documents since imports are not resolved at this
		// moment
		return createDocument(documentId, xmlElement.getTagName(), attributes, children, new ArrayList<Document>());
	}

	private static Set<Attribute> createAttributeSet(Element xmlElement) throws IllegalArgumentException {
		Set<Attribute> attributes = new HashSet<Attribute>();
		NamedNodeMap xmlAttributes = xmlElement.getAttributes();
		if (xmlAttributes != null) {
			for (int a = 0; a < xmlAttributes.getLength(); a++) {
				org.w3c.dom.Node xmlAttribute = xmlAttributes.item(a);
				attributes.add(createAttribute(xmlAttribute));
			}
		}
		return attributes;
	}

	private static Node createNode(Element xmlElement) throws IllegalArgumentException {
		Set<Attribute> attributes = createAttributeSet(xmlElement);
		List<Node> children = new ArrayList<Node>();
		NodeList xmlChildNodes = xmlElement.getChildNodes();
		for (int c = 0; c < xmlChildNodes.getLength(); c++) {
			org.w3c.dom.Node child = xmlChildNodes.item(c);
			if (child instanceof Element) {
				Node createdNode = createNode((Element) child);
				if (createdNode != null) {
					children.add(createdNode);
				}
			}
		}
		return createNode(xmlElement.getTagName(), attributes, children);
	}

}
