/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser.test.xmlparser;

import java.io.FileNotFoundException;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.parser.XMLParser;
import com.is2t.giml.xml.parser.XMLParserOptions;
import com.is2t.giml.xml.parser.exception.XMLParserException;

/**
 * Test cases of {@link XMLParser} on valid documents and imports are not resolved.
 */
public class XMLParserValidDocumentsNotResolvedImportsTest extends AbstractXMLParserValidTest {

	private static final String RESOURCE_FOLDER_PATH = "/com/is2t/giml/xml/parser/test/xmlparser/imports/valid/";

	private static final String PARSER_OPTION_XML_EXTENSION = ".xml";
	private static final String PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME = "name";
	private static final String PARSER_OPTION_IMPORT_TAGNAME = "import";
	private static final String PARSER_OPTION_IMPORT_ATTRIBUTE_NAME = "name";
	private static final boolean PARSER_OPTION_HANDLE_IMPORTS = false;

	@Override
	protected XMLParserOptions getXMLParserOptions() {
		XMLParserOptions options = new XMLParserOptions();
		options.setXMLFileExtension(PARSER_OPTION_XML_EXTENSION);
		options.setDocumentIdAttributeName(PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME);
		options.setImportTagName(PARSER_OPTION_IMPORT_TAGNAME);
		options.setImportTagAttributeName(PARSER_OPTION_IMPORT_ATTRIBUTE_NAME);
		options.setHandleImports(PARSER_OPTION_HANDLE_IMPORTS);
		return options;
	}

	@Override
	protected String getResourceFolder() {
		return RESOURCE_FOLDER_PATH;
	}

	@Test
	public void testSimpleDocument() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("simple/simple.xml")).parse();

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "simpleId");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 2);
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check first child
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check second child
		Node subnode2 = createNode("simplesubnode2", NO_ATTRIBUTE);
		Node node2 = createNode("simplenode2", NO_ATTRIBUTE, subnode2);
		Assert.assertTrue(document.getChildren().contains(node2));
	}

	@Test
	public void testSimpleDocumentsWithImport() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("simpleimport/doc1.xml")).parse();

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "doc1");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 3); // import not resolved so tag is kept
		Assert.assertEquals(document.getReferencedDocumentCount(), 0); // import not resolved
		Assert.assertTrue(document.hasChild(PARSER_OPTION_IMPORT_TAGNAME));

		// Check first child
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check second child
		Node subnode2 = createNode("simplesubnode2", NO_ATTRIBUTE);
		Node node2 = createNode("simplenode2", NO_ATTRIBUTE, subnode2);
		Assert.assertTrue(document.getChildren().contains(node2));

		// Check import child
		Set<Attribute> importNodeAttrSet = createAttributeSet("name", "doc2");
		Node importNode = createNode("import", importNodeAttrSet);
		Assert.assertTrue(document.getChildren().contains(importNode));
	}

	@Test
	public void testDocumentsWithChainImport() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("chainimports/doc1.xml")).parse();

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "doc1");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 4); // import not resolved so tag is kept
		Assert.assertEquals(document.getReferencedDocumentCount(), 0); // import not resolved
		Assert.assertTrue(document.hasChild(PARSER_OPTION_IMPORT_TAGNAME)); // import are resolved

		// Check doc1 node
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check import childs
		Set<Attribute> import1NodeAttrSet = createAttributeSet("name", "doc2");
		Node import1Node = createNode("import", import1NodeAttrSet);
		Assert.assertTrue(document.getChildren().contains(import1Node));
		Set<Attribute> import2NodeAttrSet = createAttributeSet("name", "doc3");
		Node import2Node = createNode("import", import2NodeAttrSet);
		Assert.assertTrue(document.getChildren().contains(import2Node));
		Set<Attribute> import4NodeAttrSet = createAttributeSet("name", "doc4");
		Node import4Node = createNode("import", import4NodeAttrSet);
		Assert.assertTrue(document.getChildren().contains(import4Node));
	}

}
