/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser.utils;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

/**
 * The class {@code FileUtils} contains utility methods to handle files.
 */
public class FileUtils {

	private static final String NULL_AS_STRING = "null";

	private FileUtils() {
		// utility class
	}

	/**
	 * Gets if the given argument denotes a valid file pathname.
	 *
	 * An instance of {@link File} denotes a valid file pathname if the file that it references exists and is a file
	 * (not a directory).
	 *
	 * @param file
	 *            the file to check.
	 * @return <code>true</code> if the given file is a valid file, <code>false</code> otherwise.
	 */
	public static boolean isValidFile(File file) {
		return file != null && file.exists() && file.isFile();
	}

	/**
	 * Gets if the given argument denotes a valid directory pathname.
	 *
	 * An instance of {@link File} denotes a valid directory pathname if the file that it references exists and is a
	 * directory (not a file).
	 *
	 * @param dir
	 *            the directory to check.
	 * @return <code>true</code> if the given file is a valid directory, <code>false</code> otherwise.
	 */
	public static boolean isValidDirectory(File dir) {
		return dir != null && dir.exists() && dir.isDirectory();
	}

	/**
	 * Ensures the given file is a valid file.
	 *
	 * @param file
	 *            the file to check.
	 * @throws IllegalArgumentException
	 *             if the given file does not denote a valid file pathname.
	 *
	 * @see #isValidFile(File)
	 */
	public static void checkValidFile(File file) throws IllegalArgumentException {
		if (!isValidFile(file)) {
			throw new IllegalArgumentException(String.format("'%s' is not a valid file.",
					(file != null ? file.getAbsolutePath() : NULL_AS_STRING)));
		}
	}

	/**
	 * Ensures the given file is a valid directory.
	 *
	 * @param directory
	 *            the directory to check.
	 * @throws IllegalArgumentException
	 *             if the given file does not denote a valid directory pathname.
	 *
	 * @see #isValidFile(File)
	 */
	public static void checkValidDirectory(File directory) throws IllegalArgumentException {
		if (!isValidDirectory(directory)) {
			throw new IllegalArgumentException(String.format("'%s' is not a valid directory.",
					(directory != null ? directory.getAbsolutePath() : NULL_AS_STRING)));
		}
	}

	/**
	 * Ensures the files in the directories list are valid directories.
	 *
	 * @param directories
	 *            the list of directories to check.
	 * @throws IllegalArgumentException
	 *             if a file in the list does not denote a valid directory pathname.
	 *
	 * @see #isValidFile(File)
	 */
	public static void checkValidDirectory(List<File> directories) throws IllegalArgumentException {
		if (directories == null || directories.size() == 0) {
			throw new IllegalArgumentException("No valid directories list");
		}
		for(File directory: directories){
			checkValidDirectory(directory);
		}
	}
	/**
	 * Gets the list of files in the given folder and its sub-folders, which match the given extension.
	 *
	 * @param directory
	 *            the parent directory to look for files.
	 * @param extension
	 *            the extension of file to look for.
	 * @return a possibly empty list of files that match the given extension.
	 * @throws NullPointerException
	 *             if given arguments are <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given folder does not exist or does not denote a folder pathname.
	 */
	public static List<File> listFiles(File directory, String extension)
			throws NullPointerException, IllegalArgumentException {
		if (!directory.exists() || !directory.isDirectory()) {
			throw new IllegalArgumentException("Given file does not denote an existing directory pathname");
		}
		List<File> files = new ArrayList<File>();

		File[] children = directory.listFiles();
		for (File child : children) {
			if (child.isFile()) {
				if (child.getName().endsWith(extension)) {
					files.add(child);
				}
			} else {
				// is directory
				files.addAll(listFiles(child, extension));
			}
		}

		return files;
	}

	public static List<File> listFiles(List<File> directories, String extension) {
		List<File> files = new ArrayList<File>();
		for (File directory : directories) {
			files.addAll(listFiles(directory, extension));
		}
		return files;
	}
}
