/*
 * Java
 *
 * Copyright 2009-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.java;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;

import com.is2t.java.nodes.AbstractMethodDeclaration;
import com.is2t.java.nodes.AbstractVariableDeclaration;
import com.is2t.java.nodes.AddressReference;
import com.is2t.java.nodes.Block;
import com.is2t.java.nodes.BreakStatement;
import com.is2t.java.nodes.CaseEntry;
import com.is2t.java.nodes.CatchClause;
import com.is2t.java.nodes.ClassDeclaration;
import com.is2t.java.nodes.Comment;
import com.is2t.java.nodes.CompilationUnit;
import com.is2t.java.nodes.ComposeFieldOrLocalDeclaration;
import com.is2t.java.nodes.ConstructorDeclaration;
import com.is2t.java.nodes.ContinueStatement;
import com.is2t.java.nodes.DoStatement;
import com.is2t.java.nodes.Expression;
import com.is2t.java.nodes.FieldDeclaration;
import com.is2t.java.nodes.ForStatement;
import com.is2t.java.nodes.GotoStatement;
import com.is2t.java.nodes.IfStatement;
import com.is2t.java.nodes.Initializer;
import com.is2t.java.nodes.InterfaceDeclaration;
import com.is2t.java.nodes.JavaSourceNode;
import com.is2t.java.nodes.JavadocComment;
import com.is2t.java.nodes.LabelStatement;
import com.is2t.java.nodes.LocalDeclaration;
import com.is2t.java.nodes.MethodDeclaration;
import com.is2t.java.nodes.Modifiers;
import com.is2t.java.nodes.ReturnStatement;
import com.is2t.java.nodes.SemiColonStatement;
import com.is2t.java.nodes.SimpleComment;
import com.is2t.java.nodes.Statement;
import com.is2t.java.nodes.SwitchStatement;
import com.is2t.java.nodes.SynchronizedStatement;
import com.is2t.java.nodes.ThrowStatement;
import com.is2t.java.nodes.TryStatement;
import com.is2t.java.nodes.TypeDeclaration;
import com.is2t.java.nodes.TypeReference;
import com.is2t.java.nodes.WhileStatement;
import com.is2t.java.nodes.expressions.Allocation;
import com.is2t.java.nodes.expressions.ArrayAccess;
import com.is2t.java.nodes.expressions.ArrayAllocation;
import com.is2t.java.nodes.expressions.Assignment;
import com.is2t.java.nodes.expressions.BinaryExpression;
import com.is2t.java.nodes.expressions.BooleanLiteral;
import com.is2t.java.nodes.expressions.ByteLiteral;
import com.is2t.java.nodes.expressions.CastExpression;
import com.is2t.java.nodes.expressions.CharLiteral;
import com.is2t.java.nodes.expressions.ClassLiteral;
import com.is2t.java.nodes.expressions.CompoundAssignment;
import com.is2t.java.nodes.expressions.DoubleLiteral;
import com.is2t.java.nodes.expressions.DoubleLiteralInDoubleClass;
import com.is2t.java.nodes.expressions.ExplicitConstructorInvocation;
import com.is2t.java.nodes.expressions.FieldAccess;
import com.is2t.java.nodes.expressions.FloatLiteral;
import com.is2t.java.nodes.expressions.FloatLiteralInFloatClass;
import com.is2t.java.nodes.expressions.IncrementExpression;
import com.is2t.java.nodes.expressions.IntegerLiteral;
import com.is2t.java.nodes.expressions.LocalAccess;
import com.is2t.java.nodes.expressions.LongLiteral;
import com.is2t.java.nodes.expressions.MessageSend;
import com.is2t.java.nodes.expressions.NullLiteral;
import com.is2t.java.nodes.expressions.ObjectAllocation;
import com.is2t.java.nodes.expressions.ShortLiteral;
import com.is2t.java.nodes.expressions.StringLiteral;
import com.is2t.java.nodes.expressions.TernaryExpression;
import com.is2t.java.nodes.expressions.ThisExpression;
import com.is2t.java.nodes.expressions.TypeReferenceExpression;
import com.is2t.java.nodes.expressions.UnaryExpression;


public class JavaSourceToStringGenerator extends JavaSourceGenerator{

	public static final String StrEOL = System.getProperty("line.separator");
	public static final String DotJava = ".java";

	public static final String StrBoolean 		= "boolean";
	public static final String StrByte 			= "byte";
	public static final String StrChar 			= "char";
	public static final String StrClass 		= "class";
	public static final String StrInterface 	= "interface";
	public static final String StrDouble 		= "double";
	public static final String StrExtends 		= "extends";
	public static final String StrFloat 		= "float";
	public static final String StrImplements 	= "implements";
	public static final String StrInt 			= "int";
	public static final String StrJavadocStart	= "/**";
	public static final String StrDocStart		= "/*";
	public static final String StrDocEnd		= "*/";
	public static final String StrLength		= "length";
	public static final String StrLong 			= "long";
	public static final String StrPackage 		= "package";
	public static final String StrReturn 		= "return";
	public static final String StrShort 		= "short";
	public static final String StrVoid 			= "void";
	public static final String StrBreak 		= "break";
	public static final String StrContinue		= "continue";
	public static final String StrImport 		= "import";
	public static final String StrIf			= "if";
	public static final String StrElse			= "else";
	public static final String StrNew			= "new";
	public static final String StrFor			= "for";
	public static final String StrWhile			= "while";
	public static final String StrDo			= "do";
	public static final String StrThis			= "this";
	public static final String StrSuper			= "super";
	public static final String StrCase			= "case";
	public static final String StrSwitch		= "switch";
	public static final String StrDefault		= "default";
	public static final String StrTry			= "try";
	public static final String StrFinally		= "finally";
	public static final String StrCatch			= "catch";
	public static final String StrThrow			= "throw";
	public static final String StrThrows		= "throws";
	public static final String StrStatic		= "static";
	public static final String StrSynchronized 	= "synchronized";
	public static final String StrNan			= "NaN";
	public static final String StrNegativeInfinity		= "NEGATIVE_INFINITY";
	public static final String StrPositiveInfinity		= "POSITIVE_INFINITY";

	public static final char Comma 				= ',';
	public static final char BackSlash 			= '\\';
	public static final char Dot 				= '.';
	public static final char DoubleQuote		= '\"';
	public static final char Quote				= '\'';
	public static final char Equal				= '=';
	public static final char LeftBrace			= '{';
	public static final char LeftBracket		= '[';
	public static final char LeftParenthesis	= '(';
	public static final char RightBrace			= '}';
	public static final char RightBracket		= ']';
	public static final char RightParenthesis	= ')';
	public static final char Semicolon 			= ';';
	public static final char Space 				= ' ';
	public static final char Div 				= '/';
	public static final char Star 				= '*';
	public static final char Tab 				= '\t';
	public static final char ExclamationMark	= '?';
	public static final char Colon				= ':';
	public static final char Coma				= ',';

	protected StringBuffer stream;
	private int tabs;


	private JavaSourceToStringPregenerator preGenerator;
	private boolean fullyParenthesis = true;
	private Hashtable importEntries;

	private TypeReference currentTypeReference;

	/**
	 * Without preGenerator
	 */
	public JavaSourceToStringGenerator(){

	}
	/**
	 * With preGenerator for toString presentation enhancement
	 */
	public JavaSourceToStringGenerator(JavaSourceToStringPregenerator preGenerator){
		this.preGenerator = preGenerator;
	}

	/** API
	 * return the String representation of the source node
	 */
	public String toString(JavaSourceNode node){
		initialize();
		if(node instanceof Expression){
			fullyParenthesis = false;
			((Expression)node).generateExpressionUsing(this);
		}
		else{
			// generated as a statement
			node.generateUsing(this);
		}
		return stream.toString();
	}

	private void initialize() {
		importEntries = new Hashtable();
		stream = new StringBuffer(10000);
	}
	/**
	 * API
	 * @return null if something goes wrong when generating the model (bad construction)
	 */
	public byte[] generate(CompilationUnit unit){
		initialize();
		try{ unit.generateUsing(this); }
		catch(Throwable e){
			e.printStackTrace();
			return null;
		}
		return stream.toString().getBytes();
	}

	/**
	 * API
	 * @return null if something goes wrong when generating the model (bad construction)
	 */
	public byte[] generateHeader(CompilationUnit unit){
		stream = new StringBuffer(10000);
		try{ generateFileHeader(unit); }
		catch(Throwable e){
			e.printStackTrace();
			return null;
		}
		return stream.toString().getBytes();
	}

	/**
	 * API
	 * Generate a CompilationUnit and dump it in the given output directory
	 * Packages are automatically created
	 * Example:
	 * generate(a.b.C, outputDir) => generate outputDir/a/b/C.java
	 */
	public boolean generate(CompilationUnit unit, String outputDir){
		byte[] bytes = generate(unit);
		if(bytes == null) {
			return false;
		}

		byte[] header = generateHeader( unit);

		String filename = expectedGenerateFilename(unit, outputDir);
		createDirectoriesChain(new File(filename).getParent());
		FileOutputStream fos = null;
		try{
			fos = new FileOutputStream(filename);
			fos.write( header);
			fos.flush();
			fos.write(bytes);
			fos.flush();
		}
		catch(IOException e){
			return false;
		}
		finally{
			try{ fos.close(); }
			catch(Throwable e){}
		}

		return true;
	}

	/**
	 * API
	 * Given a CompilationUnit and an output directory, return the expected filename
	 */
	public String expectedGenerateFilename(CompilationUnit unit, String outputDir){
		char[][] packageName = unit.packageName;
		StringBuffer sb = new StringBuffer(outputDir).append(File.separatorChar);
		for(int i=-1; ++i<packageName.length;){
			sb.append(packageName[i]).append(File.separatorChar);
		}
		sb.append(unit.name).append(getFileExtension());
		return sb.toString();
	}

	protected String getFileExtension() {
		return DotJava;
	}

	public void generateFileHeader(CompilationUnit node) {
		generateComment(node);
		stream.append(StrPackage).append(Space);
		generateFullyQualifiedName(node.packageName);
		stream.append(Semicolon);
		stream.append(StrEOL);
		crTab();

		// sort elements to have same order for 2 generations
		ArrayList list = new ArrayList();

		Iterator iter = importEntries.values().iterator();
		while ( iter.hasNext()) {
			StringBuffer sb = new StringBuffer();
			fullyQualifiedName(((TypeReference) iter.next()).fullyQualifiedName, sb);
			list.add( sb.toString());
		}

		Collections.sort( list);

		for (int i = list.size(); --i >=0;) {
			stream.append(StrImport).append(Space);
			stream.append( list.get( i));
			stream.append(Semicolon);
			crTab();
		}

	}

	@Override
	public void generateCompilationUnit(CompilationUnit node) {

		TypeDeclaration[] types = node.types;
		int nbTypes = types.length;
		for(int i=-1; ++i<nbTypes;){
			crTab();
			types[i].generateUsing(this);
		}
	}

	private void generateQualifiedName(char[][] name, boolean withDot) {
		int size = name.length;
		char[] classname = name[ size-1];

		TypeReference existingType = (TypeReference) importEntries.get(classname);
		TypeReference newType =  new TypeReference( name, 0);
		boolean isCurrentType = currentTypeReference != null && currentTypeReference.equals(newType);
		if (isCurrentType) {
			// NOP
		} else if (existingType != null && newType.equals(existingType)) {
			stream.append( classname);
			if (withDot) {
				stream.append(Dot);
			}
		}
		else { // an import already exists for this classname but with another package
			generateFullyQualifiedName( name);
			if (withDot) {
				stream.append(Dot);
			}
		}
	}


	private void generateFullyQualifiedName(char[][] name) {
		fullyQualifiedName( name, stream);
	}

	private void fullyQualifiedName(char[][] name, StringBuffer sb) {
		int length = name.length;
		sb.append(name[0]);
		for(int i=0; ++i<length;){
			sb.append(Dot);
			sb.append(name[i]);
		}
	}


	@Override
	public void generateMethodDeclaration( MethodDeclaration node) {
		generateComment(node);
		node.modifiers.generateUsing(this);
		node.returnType.generateUsing(this);
		stream.append(Space);
		generateAbstractMethodDeclaration( node);
	}

	@Override
	public void generateConstructorDeclaration( ConstructorDeclaration node) {
		generateComment(node);
		node.modifiers.generateUsing(this);
		generateAbstractMethodDeclaration( node);
	}

	public void generateAbstractMethodDeclaration(AbstractMethodDeclaration node) {

		stream.append(node.name).append(LeftParenthesis);
		LocalDeclaration[] parameters = node.parameters;
		int nbParameters = parameters.length;
		for(int i=-1; ++i<nbParameters;){
			LocalDeclaration parameter = parameters[i];
			if(parameter.modifiers != null){
				parameter.modifiers.generateUsing(this);
			}
			parameter.type.generateUsing(this);
			stream.append(Space).append(parameter.name);
			if(i<nbParameters-1){
				stream.append(Comma).append(Space);
			}
		}
		stream.append(RightParenthesis).append( Space);

		//throws clause
		TypeReference[] exceptionTypes =  node.exceptions;
		if ( exceptionTypes != null) {
			//almost one Exception
			stream.append( StrThrows).append(Space);
			exceptionTypes[0].generateUsing( this);
			for (int i = 1; i < exceptionTypes.length; i++) {
				stream.append( Coma).append( Space);
				exceptionTypes[i].generateUsing( this);
			}
		}

		//body for concrete method
		if( node.modifiers.isAbstract()) {
			stream.append(Semicolon);
		}
		else {
			stream.append(LeftBrace);
			++tabs;
			Statement[] statements = node.statements;
			int nbStatements = statements.length;
			for(int i=-1; ++i<nbStatements; ){
				crTab();
				statements[i].generateUsing(this);
			}
			--tabs;
			crTab();
			stream.append(RightBrace);
		}
	}

	private static void createDirectoriesChain(String dirname){
		//creates all the directories that need to be create in order
		//for the dirname to be legitimate
		java.io.File dir;
		if ( (dir = new java.io.File(dirname) ).exists()) {
			return;
		}
		String parent = dir.getParent();
		if (parent != null) {
			createDirectoriesChain(parent);
		}
		dir.mkdir();
	}

	public void crTab(){
		stream.append(StrEOL);
		for(int i=tabs; --i>=0;){
			stream.append(Tab);
		}
	}


	/**
	 * comments may be null
	 */
	private void generateComments(Comment[] comments){
		if(comments == null) {
			return;
		}
		int nbComments = comments.length;
		for(int i=-1; ++i<nbComments;){
			comments[i].generateUsing(this);
		}
	}


	private void addImplementors( TypeDeclaration node, String keyword) {
		TypeReference[] interfaces = node.interfaces;
		if(interfaces.length > 0){
			stream.append(Space).append( keyword).append(Space);
			int nbInterfaces = interfaces.length;
			for(int i=-1; ++i<nbInterfaces;){
				if(i>0){
					stream.append(Comma).append(Space);
				}
				interfaces[i].generateUsing(this);
			}
		}
	}

	private void addFieldDeclaration( TypeDeclaration node){
		Statement[] fields = node.fieldsOrInitializers;
		if(fields.length > 0 ){
			int nbFields = fields.length;
			//keep same order as source !
			for(int i=0; i < nbFields; i++){
				crTab();
				fields[i].generateUsing(this);
			}
			crTab();// skip a line between field and next element
		}
	}

	private void addConstructors( ClassDeclaration node){
		ConstructorDeclaration[] constructors = node.constructors;
		if(constructors.length > 0){
			int nbConstructors = constructors.length;
			for(int i=-1; ++i<nbConstructors;){
				crTab();
				constructors[i].generateUsing(this);
			}
			crTab(); // skip a line between constructor and next element
		}
	}

	private void addMethods( TypeDeclaration node){
		MethodDeclaration[] methods = node.methods;
		if(methods.length > 0){
			int nbMethods = methods.length;
			for(int i=-1; ++i<nbMethods;){
				crTab();
				methods[i].generateUsing(this);
			}
			crTab(); // skip a line between method and next element
		}
	}


	@Override
	public void generateClassDeclaration(ClassDeclaration node){
		currentTypeReference = new TypeReference(node.fullyQualifiedName, 0);
		generateComments(node.comments);
		node.modifiers.generateUsing(this);
		stream.append(StrClass).append(Space).append(node.simpleName());
		if(node.superType != null){
			stream.append(Space).append(StrExtends).append(Space);
			node.superType.generateUsing(this);
		}

		addImplementors( node, StrImplements);

		stream.append(LeftBrace);
		crTab(); // skip a line between classdeclaration and first element

		++tabs;

		addFieldDeclaration( node);
		addConstructors( node);
		addMethods( node);

		--tabs;

		crTab();
		stream.append(RightBrace);
		currentTypeReference = null;
	}

	@Override
	public void generateInterfaceDeclaration(InterfaceDeclaration node){
		generateComments(node.comments);
		// no need abstract in interface declaration
		node.modifiers.generateUsing(this);
		stream.append(StrInterface).append(Space).append(node.simpleName());

		addImplementors( node, StrExtends);

		stream.append(LeftBrace);
		crTab();

		++tabs;

		addFieldDeclaration( node);
		addMethods( node);

		--tabs;

		crTab();
		stream.append(RightBrace);
	}

	@Override
	public void generateExpression(Expression node) {
		fullyParenthesis=false;

		node.generateExpressionUsing(this);
		stream.append(Semicolon);
		generateComment(node);
	}

	@Override
	public void generateFieldAccess(FieldAccess node) {
		fullyParenthesis = true;
		if(node.receiver != null){
			node.receiver.generateExpressionUsing(this);
			stream.append(Dot);
		}
		else {
			//static field
			generateQualifiedName(node.declarationFullyQualifiedName, true);
		}
		if(node.isArrayLength()){
			stream.append(StrLength);
		}
		else{
			stream.append(node.name);
		}
		generateComment( node);

	}

	@Override
	public void generateFieldDeclaration(FieldDeclaration node) {
		generateComment(node);
		generateAbstractVariableDeclaration(node);
	}

	public void generateAbstractVariableDeclaration(AbstractVariableDeclaration node){
		node.modifiers.generateUsing(this);
		node.type.generateUsing(this);
		stream.append(Space);
		generatePartOfVariableDeclaration( node);
		stream.append(Semicolon);
	}



	private void generatePartOfVariableDeclaration(AbstractVariableDeclaration node){

		stream.append(node.name);
		if(node.initExp != null){
			stream.append(Space).append(Equal).append(Space);
			fullyParenthesis = false;
			node.initExp.generateExpressionUsing(this);
		}

	}

	@Override
	public void generateLocalAccess(LocalAccess node) {
		stream.append(node.declaration.name);
		generateComment(node);
	}

	@Override
	public void generateLocalDeclaration(LocalDeclaration node) {
		generateAbstractVariableDeclaration(node);
		generateComment(node);
	}

	@Override
	public void generateTypeReference(TypeReference node) {
		generateTypeReference( node, true);
		generateComment(node);
	}

	public void generateComment( JavaSourceNode node ) {
		if ( node.comment != null) {
			node.comment.generateUsing( this);
		}
	}

	public void generateTypeReference(TypeReference node, boolean withDim) {
		if(node.isBasetypeOrArrayOfBasetype()){
			String typeStr;
			switch(node.id){
			case TypeReference.T_BOOLEAN : 	typeStr = StrBoolean; break;
			case TypeReference.T_BYTE : 	typeStr = StrByte; break;
			case TypeReference.T_SHORT : 	typeStr = StrShort; break;
			case TypeReference.T_CHAR : 	typeStr = StrChar; break;
			case TypeReference.T_INT : 		typeStr = StrInt; break;
			case TypeReference.T_LONG : 	typeStr = StrLong; break;
			case TypeReference.T_FLOAT :	typeStr = StrFloat; break;
			case TypeReference.T_DOUBLE : 	typeStr = StrDouble; break;
			case TypeReference.T_VOID : 	typeStr = StrVoid; break;
			default:throw new IllegalArgumentException();
			}
			stream.append(typeStr);
		}
		else {
			generateQualifiedName(node.fullyQualifiedName, false);
		}

		if ( withDim) {
			for(int i=node.dimensions; --i>=0;){
				stream.append(LeftBracket).append(getDimQualifier(node.getDimQualifier(i))).append(RightBracket);
			}
		}
	}

	/**
	 * Java has no dimension qualifier
	 */
	public String getDimQualifier(int value){
		if(value == TypeReference.DimNoQualifier) {
			return "";
		} else {
			throw new AssertionError();
		}
	}

	@Override
	public void generateModifiers(Modifiers node){
		generateModifiers(node, false); // by default
	}

	public void generateModifiers(Modifiers node, boolean isSuperAllowed) {

		int bits = node.bits;
		boolean notFirst = false;
		if ((bits & Modifiers.AccDeprecated) != 0 ){
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("/**@deprecated*/");
		}

		if ((bits & Modifiers.AccPublic) != 0 ){
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("public");
		}

		if ((bits & Modifiers.AccPrivate) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("private");
		}

		if ((bits & Modifiers.AccProtected) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("protected");
		}

		if ((bits & Modifiers.AccStatic) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("static");
		}

		if ((bits & Modifiers.AccFinal) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("final");
		}

		if ((bits & Modifiers.AccSuper) != 0 && isSuperAllowed) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("super");
		}

		if ((bits & Modifiers.AccSynchronized) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append(StrSynchronized);
		}

		if ((bits & Modifiers.AccVolatile) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("volatile");}

		if ((bits & Modifiers.AccTransient) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("transient");
		}

		if ((bits & Modifiers.AccNative) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("native");
		}

		if ((bits & Modifiers.AccInterface) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("interface");
		}

		if ((bits & Modifiers.AccAbstract) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("abstract");
		}

		if ((bits & Modifiers.AccStrictfp) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("strictfp");
		}

		if ((bits & Modifiers.AccUpper) != 0 ) {
			if (notFirst) {
				stream.append(' ');
			}
			notFirst = true;
			stream.append("upper");
		}

		if ( notFirst) {
			stream.append(Space);
		}
		generateComment( node);

	}

	@Override
	public void generateMessageSend(MessageSend node) {
		if(node.receiver != null){
			fullyParenthesis = true;
			node.receiver.generateExpressionUsing(this);
			stream.append(Dot);
		}
		else if (node.declarationTypeName != null && node.declarationTypeName.length != 0) {
			// no receiver : static method is generated with class.method
			// except if class == current class
			// NOTE : node.declarationTypeName[0].length != 0 for special case of a MessageSend in a MethodAllocation
			//   -> see IceteaToJavaGenerator.generateMethodAllocation()
			generateQualifiedName(node.declarationTypeName, true);
		}


		stream.append(node.name).append(LeftParenthesis);
		Expression[] arguments = node.arguments;
		int nbArguments = arguments.length;
		for(int i=-1; ++i<nbArguments;){
			fullyParenthesis=false;
			arguments[i].generateExpressionUsing(this);
			if(i<nbArguments-1){
				stream.append(Comma).append(Space);
			}
		}
		stream.append(RightParenthesis);
		generateComment( node);

	}

	@Override
	public void generateReturnStatement(ReturnStatement node) {
		//		if ( node.previousStatement != null) {
		//			//TODO find a most elegant solution for return on a void method with postcondition ...
		//			node.previousStatement.generateUsing( this);
		//			stream.append(Space);
		//		}

		stream.append(StrReturn);
		if(node.result != null){
			stream.append(Space);
			fullyParenthesis =false;
			node.result.generateExpressionUsing(this);
		}
		stream.append(Semicolon);
		generateComment( node);

	}

	@Override
	public void generateStringLiteral(StringLiteral node) {

		stream.append(DoubleQuote);
		char[] token = node.value;
		for(int i=-1,length=token.length; ++i <length;){
			generateCharacter(token[i]);
		}
		stream.append(DoubleQuote);
		generateComment( node);

	}

	@Override
	public void generateJavadocComment(JavadocComment node) {
		stream.append(StrJavadocStart);
		char[][] lines = node.lines;
		int nbLines = lines.length;
		for(int linePtr=-1; ++linePtr<nbLines;){
			char[] line = lines[linePtr];
			crTab();
			stream.append(Space).append(Star).append(Space);
			for(int i=-1,length=line.length; ++i <length;){
				generateCharacter(line[i]);
			}
		}
		crTab();
		stream.append(Space).append(StrDocEnd);
		crTab();
	}

	@Override
	public void generateAssignment(Assignment node) {
		boolean bool = fullyParenthesis ;
		if (bool) {
			stream.append(LeftParenthesis);
		}

		node.left.generateExpressionUsing(this);
		stream.append(Space).append(Equal).append(Space);
		fullyParenthesis = false;
		node.right.generateExpressionUsing(this);

		if (bool) {
			stream.append(RightParenthesis);
		}

		generateComment( node);
	}

	@Override
	public void generateBreakStatement(BreakStatement node) {
		stream.append(StrBreak);
		generateGotoStatement( node);
	}

	public void generateGotoStatement(GotoStatement node) {
		if ( node.label != null) {
			stream.append(Space);
			stream.append( node.label);
		}
		stream.append(Semicolon);
		generateComment( node);
	}


	@Override
	public void generateClassLiteral(ClassLiteral classLiteral) {
		generateTypeReference( classLiteral.type);
		generateComment( classLiteral);
	}

	@Override
	public void generateIfStatement(IfStatement ifStatement) {
		stream.append(StrIf).append(LeftParenthesis);
		fullyParenthesis=false;
		ifStatement.condition.generateExpressionUsing( this);
		stream.append(RightParenthesis).append(Space);
		generateComment(ifStatement);

		Statement branch = ifStatement.ifTrue;
		boolean simpleStatement;
		if ( branch != null) {
			simpleStatement = !(branch instanceof Block);
			if ( simpleStatement) {
				tabs ++;
				crTab();
			}

			branch.generateUsing( this);

			if ( simpleStatement) {
				tabs --;
				crTab();
			}

		}

		branch = ifStatement.ifFalse;
		if ( branch != null) {
			stream.append(StrElse).append(Space);
			simpleStatement = ! (branch instanceof Block);
			if ( simpleStatement) {
				tabs ++;
				crTab();
			}

			branch.generateUsing( this);

			if ( simpleStatement) {
				tabs --;
				crTab();
			}
		}


	}


	@Override
	public void generateByteLiteral(ByteLiteral literal) {
		stream.append(LeftParenthesis);
		stream.append(StrByte);
		stream.append(RightParenthesis);
		stream.append(literal.value);
		generateComment(literal);

	}

	@Override
	public void generateCharLiteral(CharLiteral literal) {
		stream.append(Quote);
		generateCharacter(literal.value);
		stream.append(Quote);
		generateComment(literal);
	}

	private void generateCharacter( char value) {
		switch (value) {
		case '\b' : stream.append("\\b"); break;
		case '\t' : stream.append("\\t"); break;
		case '\n' : stream.append("\\n"); break;
		case '\f' : stream.append("\\f"); break;
		case '\r' : stream.append("\\r"); break;
		case '\\' : stream.append("\\\\"); break;
		case '\'' : stream.append("\\\'"); break;
		case '\"' : stream.append("\\\""); break;
		default :
			int i = value&0xFFFF;
			if ((i<32) || (i>127))
			{
				stream.append("\\u");
				if (i<16) {
					stream.append("0");
				}
				if (i<(16*16)) {
					stream.append("0");
				}
				if (i<(16*16*16)) {
					stream.append("0");
				}
				stream.append(Integer.toHexString(i));
			}
			else
			{
				stream.append(value);
			}
		}
	}

	@Override
	public void generateDoubleLiteral(DoubleLiteral literal) {
		double d = literal.value;
		Double nodeValue = new Double( d) ;
		if ( nodeValue.equals( new Double( Double.NEGATIVE_INFINITY))) {
			stream.append("Double.NEGATIVE_INFINITY" );
		} else if ( nodeValue.equals( new Double( Double.POSITIVE_INFINITY))) {
			stream.append("Double.POSITIVE_INFINITY" );
		} else if ( nodeValue.equals( new Double( Double.NaN))) {
			stream.append("Double.NaN" );
		} else {
			stream.append(d);
			stream.append('d');
			generateComment(literal);
		}
	}

	@Override
	public void generateFloatLiteral(FloatLiteral literal) {
		float f = literal.value;
		Float nodeValue = new Float( f) ;
		if ( nodeValue.equals( new Float( Float.NEGATIVE_INFINITY))) {
			stream.append("Float.NEGATIVE_INFINITY" );
		} else if ( nodeValue.equals( new Float( Float.POSITIVE_INFINITY))) {
			stream.append("Float.POSITIVE_INFINITY" );
		} else if ( nodeValue.equals( new Float( Float.NaN))) {
			stream.append("Float.NaN" );
		} else {
			stream.append(f);
			stream.append('f');
			generateComment(literal);
		}
	}


	@Override
	public void generateIntegerLiteral(IntegerLiteral literal) {
		stream.append(literal.value);
		generateComment(literal);
	}

	@Override
	public void generateLongLiteral(LongLiteral literal) {
		stream.append(literal.value);
		stream.append('l');
		generateComment(literal);

	}

	@Override
	public void generateShortLiteral(ShortLiteral literal) {
		stream.append(LeftParenthesis);
		stream.append(StrShort);
		stream.append(RightParenthesis);
		stream.append(literal.value);
		generateComment(literal);
	}

	@Override
	public void generateBinaryExpression(BinaryExpression node) {
		boolean bool = fullyParenthesis;
		fullyParenthesis=true;
		if (bool) {
			stream.append( LeftParenthesis);
		}
		node.left.generateExpressionUsing(this);

		stream.append( Space);
		Expression.operatorToString( node.operator, stream);
		stream.append( Space);

		fullyParenthesis=true;
		node.right.generateExpressionUsing(this);
		if (bool) {
			stream.append( RightParenthesis);
		}

		generateComment(node);

	}

	@Override
	public void generateUnaryExpression(UnaryExpression node) {
		boolean bool = fullyParenthesis;
		fullyParenthesis = true;
		if (bool) {
			stream.append( LeftParenthesis);
		}

		Expression.operatorToString( node.operator, stream);

		node.operand.generateExpressionUsing(this);
		if (bool) {
			stream.append( RightParenthesis);
		}

		generateComment(node);

	}

	@Override
	public void generateIncrementExpression(IncrementExpression node) {
		boolean bool = fullyParenthesis;
		fullyParenthesis = true;
		if (bool) {
			stream.append( LeftParenthesis);
		}

		if ( node.mode == IncrementExpression.Prefix) {
			Expression.operatorToString( node.operator, stream);
		}
		node.operand.generateExpressionUsing(this);
		if ( node.mode == IncrementExpression.Postfix) {
			Expression.operatorToString( node.operator, stream);
		}

		if (bool) {
			stream.append( RightParenthesis);
		}

		generateComment(node);

	}

	@Override
	public void generateObjectAllocation(ObjectAllocation node) {
		stream.append(getAllocationKeyword(node));
		stream.append( Space);

		node.type.generateUsing(this);
		stream.append(LeftParenthesis);

		if ( node.innerClass == null ){
			int size = node.parameters.length;
			if ( size > 0) {
				fullyParenthesis = false;
				stream.append(Space);
				node.parameters[0].generateExpressionUsing(this);
			}

			for ( int i = 0 ; ++i< size;) {
				fullyParenthesis = false;
				stream.append(Comma).append(Space);
				node.parameters[i].generateExpressionUsing(this);
			}
		}

		stream.append(RightParenthesis);

		if ( node.innerClass != null ) {
			generateInnerClass( node.innerClass);
		}
		generateComment(node);

	}

	protected String getAllocationKeyword(Allocation node) {
		return StrNew;
	}

	private void generateInnerClass(TypeDeclaration node){
		stream.append(LeftBrace);
		++tabs;
		addFieldDeclaration( node);
		addMethods( node);
		--tabs;
		stream.append(RightBrace);
	}

	@Override
	public void generateArrayAllocation( ArrayAllocation node) {

		stream.append(getAllocationKeyword(node));
		stream.append( Space);

		//the type HAS its dimensions set==> do not print them twice (false as second argument)
		generateTypeReference(node.type,false);

		Expression[] lstExp = node.parameters;
		int dimension =  node.getDimensions();

		for ( int i = -1 ; ++i< dimension;) {
			stream.append(LeftBracket);
			if ( lstExp != null &&  lstExp[i] != null) {
				lstExp[i].generateExpressionUsing(this);
			}

			stream.append(RightBracket);
		}

		//if initial values
		lstExp = node.initialValues;
		if ( lstExp != null) {
			stream.append(Space).append(LeftBrace);
			int initSize = lstExp.length;

			for ( int i = -1 ; ++i< initSize;) {
				stream.append(Space);
				lstExp[i].generateExpressionUsing(this);
				if ( i != initSize -1) {
					stream.append(Comma);
				}
			}
			stream.append(RightBrace);

		}

		generateComment(node);

	}

	@Override
	public void generateArrayAccess(ArrayAccess node) {
		boolean bool = fullyParenthesis;
		fullyParenthesis = true;
		if (bool) {
			stream.append( LeftParenthesis);
		}

		node.receiver.generateExpressionUsing( this);

		if (bool) {
			stream.append( RightParenthesis);
		}

		stream.append( LeftBracket);
		node.index.generateExpressionUsing( this);
		stream.append( RightBracket);

		generateComment(node);

	}

	@Override
	public void generateBlock(Block node) {

		int size;
		if ( node.statements != null) {
			size = node.statements.length;
		} else {
			size =0;
		}

		stream.append( LeftBrace);
		generateComment(node);

		if ( size != 0) {
			++tabs;
			crTab();
			size--;
			for (int i = -1; ++i< size;){
				fullyParenthesis=false;
				node.statements[i].generateUsing(this);
				crTab();
			}
			//last one
			fullyParenthesis=false;
			node.statements[size].generateUsing(this);
			--tabs;
			crTab();
		}
		stream.append( RightBrace);
		crTab();
	}

	@Override
	public void generateContinueStatement(ContinueStatement node) {
		stream.append(StrContinue);
		generateGotoStatement( node);
	}

	@Override
	public void generateForStatement(ForStatement node) {
		stream.append(StrFor);
		stream.append(LeftParenthesis);
		stream.append(Space);
		generateComment(node);

		//initialization
		Expression[] initialization = node.initialization;
		int size = initialization.length;
		for (int i = 0; i < size; i++) {
			fullyParenthesis = false;
			if ( i >0) {
				stream.append(Comma).append(Space);
			}
			initialization[i].generateExpressionUsing(this);
		}

		// or declaration
		AbstractVariableDeclaration[] declarations = node.declarations;
		size = declarations.length;
		if ( size >0) {
			//we put the dimension after local names while
			//they may not have all the same dimension ... :-(
			//for (int i[],j,k[][];....
			//This next code do not follow the yoyo design scheme
			LocalDeclaration local;
			generateTypeReference( ((LocalDeclaration)(node.declarations[0])).type, false);
			stream.append(Space);

			for( int i=0; i< size; i++) {
				if (i>0) {
					stream.append(Comma).append(Space);
				}
				local = (LocalDeclaration) (node.declarations[i]);
				stream.append(local.name);
				for (int di=local.type.dimensions;--di>=0;) {
					stream.append("[]");
				}
				stream.append(' ');
				if (local.initExp!=null) {
					stream.append("= ");
					fullyParenthesis = false;
					local.initExp.generateExpressionUsing(this);
				}
			}
		}
		stream.append(Semicolon).append(Space);


		//condition
		if ( node.condition != null) {
			fullyParenthesis = false;
			node.condition.generateUsing(this);
		}
		stream.append(Space);

		//updates
		Expression[] updates = node.updates;
		size = updates.length;
		for (int i = 0; i < size; i++) {
			fullyParenthesis =false;
			if ( i >0) {
				stream.append(Comma).append(Space);
			}
			fullyParenthesis = false;
			updates[i].generateExpressionUsing(this);
		}

		stream.append(RightParenthesis);
		crTab();
		//statement
		node.statement.generateUsing(this);

	}
	@Override
	public void generateDoStatement(DoStatement node) {
		crTab();
		stream.append(StrDo);
		generateComment(node);

		node.statement.generateUsing(this);
		stream.append(StrWhile).append(LeftParenthesis);

		fullyParenthesis=false;
		node.condition.generateExpressionUsing(this);
		stream.append(RightParenthesis);
		stream.append(Semicolon);
	}

	@Override
	public void generateWhileStatement(WhileStatement node) {
		crTab();
		stream.append(StrWhile).append(Space).append(LeftParenthesis);
		generateComment(node);
		fullyParenthesis=false;
		node.condition.generateExpressionUsing(this);
		stream.append(RightParenthesis);
		node.statement.generateUsing(this);
	}

	@Override
	public void generateCastExpression(CastExpression node) {
		boolean bool = fullyParenthesis;
		if (bool) {
			stream.append( LeftParenthesis);
		}

		fullyParenthesis = false;
		stream.append(LeftParenthesis);
		node.toType.generateUsing( this);
		stream.append(RightParenthesis);

		fullyParenthesis = true;
		node.expression.generateExpressionUsing(this);
		if (bool) {
			stream.append( RightParenthesis);
		}


	}

	@Override
	public void generateSimpleComment(SimpleComment node) {
		stream.append(StrDocStart);
		//remove first '\n' in the comment
		int i= 0;
		while ( node.commentValue[i] == '\n' ) {
			i++;
		}
		stream.append( node.commentValue, i, node.commentValue.length - i);
		stream.append(StrDocEnd);

	}

	@Override
	public void generateThisExpression(ThisExpression node) {
		stream.append( node.isSuper ? StrSuper : StrThis );
		generateComment(node);
	}

	@Override
	public void generateStructReference(AddressReference structReference) {
		// this node should be generated as int ( instead of struct type)
		structReference.id= TypeReference.T_INT;
		generateTypeReference( structReference, false);
		structReference.id= TypeReference.T_REF;
	}

	@Override
	public void generateCompoundAssignment(CompoundAssignment node) {

		boolean bool = fullyParenthesis ;
		if (bool) {
			stream.append('(');
		}

		node.left.generateExpressionUsing(this);

		stream.append( Space);
		Expression.operatorToString( node.operator, stream);
		stream.append( Space);

		fullyParenthesis = false;
		node.right.generateExpressionUsing(this);
		if (bool) {
			stream.append(')');
		}
		generateComment(node);
	}

	@Override
	public void generateLabelStatement( LabelStatement node){
		crTab();
		stream.append(node.label);
		stream.append(" : ");
		crTab();
		node.blockStatement.generateUsing(this);
		generateComment(node);

	}
	@Override
	public void generateBooleanLiteral(BooleanLiteral node) {
		stream.append(node.getValue());
		generateComment(node);

	}

	@Override
	public void generateNullLiteral(NullLiteral node) {
		stream.append(node.getValue());
		generateComment(node);
	}
	@Override
	public void genarateTernaryExpression(TernaryExpression node) {
		boolean bool = fullyParenthesis;
		if (fullyParenthesis) {
			stream.append(LeftParenthesis);
		}
		fullyParenthesis = true;

		node.condition.generateExpressionUsing(this);
		stream.append(ExclamationMark);

		fullyParenthesis = false;
		node.ifTrue.generateExpressionUsing(this);
		stream.append(Colon);

		fullyParenthesis=false;
		node.ifFalse.generateExpressionUsing(this);
		if (bool) {
			stream.append(RightParenthesis);
		}
	}

	@Override
	public void generateSwitchStatement(SwitchStatement node) {
		crTab();
		int save =tabs;
		stream.append(StrSwitch).append(LeftParenthesis).append(Space);
		fullyParenthesis = false;
		node.condition.generateExpressionUsing(this);
		stream.append(RightParenthesis).append(LeftBrace);
		generateComment(node);
		for (int i=0; i<node.statements.length;i++) {
			tabs = save +2;
			crTab();
			node.statements[i].generateUsing( this);
		}
		tabs = save;
		crTab();
		stream.append(RightBrace);
	}

	@Override
	public void generateCaseSatement(CaseEntry node) {
		tabs--;
		crTab();
		if ( node.isDefault) {
			stream.append( StrDefault);
		}
		else {
			stream.append( StrCase).append(Space);
			fullyParenthesis = false;
			node.expression.generateExpressionUsing(this);
		}
		stream.append(Space).append(Colon).append(Space);
		tabs++;
	}

	@Override
	public void generateComposeFieldOrLocalDeclaration( ComposeFieldOrLocalDeclaration node) {
		int size = node.declarations.length;
		generateAbstractVariableDeclaration(node.declarations[0]);
		//remove the SemiColon generated by generateAbstractVariableDeclaration()
		stream.deleteCharAt(stream.length()-1);

		for (int i = 0; ++i < size;) {
			stream.append(Coma).append(Space);
			generatePartOfVariableDeclaration(node.declarations[i]);
		}
		stream.append(Semicolon);
	}

	@Override
	public void generateCatchClause(CatchClause node) {

		crTab();
		stream.append( StrCatch).append(LeftParenthesis);
		node.exception.generateUsing( this);
		//remove last char = ;
		stream.deleteCharAt( stream.length()-1);
		stream.append( RightParenthesis);
		node.catchBlock.generateUsing(this);
	}

	@Override
	public void generateTryStatement(TryStatement node) {
		int index=tabs;
		crTab();
		stream.append(StrTry);
		node.tryBlock.generateUsing(this);

		for (int i=0; i<node.catches.length; i++) {
			tabs=index;
			node.catches[i].generateUsing(this);
		}

		if (node.finallyBlock != null) {
			tabs=index;
			stream.append(StrFinally);
			node.finallyBlock.generateUsing(this);
		}
		else if (node.catches.length == 0){
			//both catches and finally are null !.... pathological case !
			tabs = index;
			stream.append(StrFinally).append(LeftBrace).append(RightBrace);
		}
	}
	@Override
	public void generateThrowStatement(ThrowStatement node) {
		crTab();
		stream.append(StrThrow).append(Space);
		fullyParenthesis = false;
		node.expression.generateExpressionUsing(this);
		stream.append(Semicolon);

	}
	@Override
	public void generateSemiColon(SemiColonStatement gen) {
		stream.append( Semicolon);

	}
	@Override
	public void generateInitializerUsing(Initializer initializer) {
		stream.append( StrStatic).append(Space);
		initializer.initStatement.generateUsing( this);
	}
	@Override
	public void generateExplicitConstructorInvocation( ExplicitConstructorInvocation node) {
		if ( node.qualification != null) {
			node.qualification.generateExpressionUsing( this);
			stream.append(Dot);
		}

		if ( node.isSuper) {
			stream.append( StrSuper);
		} else {
			stream.append( StrThis);
		}

		stream.append( LeftParenthesis);

		Expression[] args = node.arguments;
		int size = args.length;
		if ( size > 0) {
			args[0].generateExpressionUsing( this);
		}

		for (int i = 0; ++i < size; ) {
			stream.append( Comma).append( Space);
			args[i].generateExpressionUsing( this);
		}

		stream.append( RightParenthesis);
	}
	@Override
	public void generateDoubleLiteralInDoubleClass( DoubleLiteralInDoubleClass node) {
		Double d = node.value;
		if ( d.equals(new Double(Double.NaN))) {
			stream.append( "0d/0d");
		} else if ( d.equals(new Double(Double.NEGATIVE_INFINITY))) {
			stream.append( "-1d/0d");
		} else if ( d.equals(new Double(Double.POSITIVE_INFINITY))) {
			stream.append( "1d/0d");
		} else {
			stream.append(d);
			stream.append('d');
			generateComment(node);
		}

	}

	@Override
	public void generateFloatLiteralInFloatClass( FloatLiteralInFloatClass node) {
		Float f = node.value ;

		if ( f.equals(new Float(Float.NaN))) {
			stream.append( "0f/0f");
		} else if ( f.equals(new Float(Float.NEGATIVE_INFINITY))) {
			stream.append( "-1f/0f");
		} else if ( f.equals(new Float(Float.POSITIVE_INFINITY))) {
			stream.append( "1f/0f");
		} else {
			stream.append(f);
			stream.append('f');
			generateComment(node);
		}

	}

	@Override
	public void generateSynchronizedStatement(SynchronizedStatement node) {
		stream.append(StrSynchronized).append(LeftParenthesis);
		node.monitor.generateExpressionUsing(this);
		stream.append(RightParenthesis);
		node.block.generateUsing(this);
	}

	@Override
	public void generateTypeReferenceExpression(TypeReferenceExpression node) {
		node.type.generateUsing(this);
	}

	/**
	 * Sets the currentTypeReference. Allows to simplify the generated code by
	 * removing unnecessary qualified names
	 *
	 * @param currentTypeReference
	 *            the currentTypeReference to set.
	 */
	public void setCurrentTypeReference(TypeReference currentTypeReference) {
		this.currentTypeReference = currentTypeReference;
	}
}
