/*
 * Java
 *
 * Copyright 2010-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.is2t.microbsp.microui.natives;

import com.is2t.vm.support.io.MemoryInputStream;

import ej.annotation.Nullable;
import ej.microui.MicroUIException;
import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;
import ej.microui.display.SImageMetadata;

public class NSystemDisplay {

	/*
	 * Native error constants
	 *
	 * WARNING copied in microui-extended
	 */

	/**
	 * no error
	 */
	public static final int NO_ERROR = 0;

	/**
	 * Native side cannot decode the file
	 */
	public static final int INVALID_TYPE = -1;

	/**
	 * oom
	 */
	public static final int OUT_OF_MEMORY = -2;

	/**
	 * File content is invalid
	 */
	public static final int INVALID_FILE = -3;

	/**
	 * File not found
	 */
	public static final int INVALID_PATH = -4;

	/**
	 * File has been closed
	 */
	public static final int FILE_CLOSED = -6;

	/**
	 * Anonymous error
	 */
	public static final int NOK = -9;

	/**
	 * Used to open image in default RAW format (default image format)
	 */
	public static final int DEFAULT_FORMAT = -1;

	/**
	 * invalid event to identify a receiver (event type 0 is "COMMAND")
	 */
	public static final int NO_END_OF_FLUSH_EVENT = 0;

	/**
	 * Suffix to apply on image path to load a RAW image.
	 */
	public static final String IMAGE_RAW_SUFFIX = "_raw"; //$NON-NLS-1$

	private NSystemDisplay() {
		// not entry point
	}

	public static native int initAndGetNextGraphicsContextId(byte[] sd, int sdSize);

	public static native int setTraceGroupID(int groupID, int eventID);

	public static native int getNbBitsPerPixel();

	public static native boolean isColor();

	public static native int getNumberOfColors();

	public static native boolean isDoubleBuffered();

	public static native boolean hasBacklight();

	/**
	 * LCD manipulations
	 */
	public static native void setContrast(int contrast);

	public static native int getContrast();

	public static native void setBacklight(int backlight);

	public static native int getBacklight();

	/**
	 * LCD offscreens
	 */
	public static native int openOffScreen(int width, int height, byte[] dstSD);

	public static native void closeOffScreen(byte[] sd);

	public static native void flushScreen(byte[] sd, boolean force);

	public static native void waitFlush(byte[] sd);

	/**
	 * LCD rendering
	 */

	// fonts
	public static native int drawChars(byte[] sd, int fontIdentifier, int x, int y, char[] chars, int offset,
			int length, int charsWidth, int[] textIndexes, byte[] externalFontHeap);

	public static native int drawDeformedChars(byte[] sd, int fontIdentifier, int x, int y, char[] chars, int offset,
			int length, int charsWidth, float xRatio, float yRatio, int[] textIndexes, byte[] externalFontHeap);

	public static native boolean isTransparent(byte[] sd);

	/**
	 * LCD read
	 */
	public static native int readPixel(byte[] sd, int x, int y);

	public static native void getARGB(byte[] srcSd, int[] rgbData, int offset, int scanlength, int x, int y, int width,
			int height);

	/**
	 * Side-effects on drawings
	 */
	public static native int getDisplayColor(int color);// alpha must be 0

	/**
	 * This method MUST BE called only by {@link GraphicsContext#resetClip()}
	 * <p>
	 * Reset the current native clip to its maximum bounds
	 */
	public static native void resetClip(byte[] sd);

	/**
	 * This method MUST BE called only by
	 * {@link GraphicsContext#resetClip(int, int, int, int)} !!!
	 * <p>
	 * Set a clip smaller or equal than current clip (update current native clip)
	 */
	public static native void setClip(byte[] sd, int x, int y, int width, int height);

	/**
	 * This method MUST BE called only by
	 * {@link GraphicsContext#setClip(int, int, int, int)} !!!
	 * <p>
	 * Set a clip smaller or equal than given wrapper clip (replace current native
	 * clip without checking its bounds (checked by caller))
	 */
	public static native void setSubClip(byte[] sd, int x, int y, int width, int height);

	public static native void setBackgroundColor(byte[] sd, int color, boolean active);

	/**
	 * @param dstSD
	 *            null to check image path only, an {@link SImageMetadata} byte
	 *            array to load the image
	 * @return an error code (see Image error codes)
	 */
	public static native int loadImageFromPath(byte[] path, int length, boolean dynamic, boolean externalResource,
			boolean rawAndByteAddressableRequired, int format, byte[] dstSD);

	public static native int loadImageFromBytes(byte[] data, int offset, int length, int format, byte[] dstSD);

	public static native int loadImageFromMIS(MemoryInputStream stream, int size, int format, byte[] dstSD);

	/**
	 * Fonts
	 */
	public static native int addInternalFont(byte[] path, int pathLength, byte[] fontData);

	public static native int getFontIdentifiers(int fontIdentifier, int[] identifiers, boolean external);

	public static native int charsWidth(int fontIdentifier, char[] chars, int offset, int length, int[] textIndexes,
			byte[] externalFontHeap);

	public static native void closeFont(int fontIdentifier, boolean external);

	/**
	 * Gives the next "end of flush" event to add in pump queue at the end of next
	 * fluh.
	 *
	 * @param event
	 *            the event to add into the pump queue
	 */
	public static native void addEndOfFlushEvent(int event);

	/**
	 * Links an existing native image (offscreen which points to a buffer located in
	 * images heap or in flash).
	 *
	 * @param sd
	 *            new image data to fill with native image
	 * @param imageReference
	 *            native image reference
	 * @param freed
	 *            true if original image (native image) has to be freed or not (this
	 *            peace of information was only available in object {@link Image}
	 *            but this object has been garbaged collected).
	 *
	 * @deprecated Kept for backward compatibility (widget2 over UI3). Task
	 *             M0092MEJAUI-2065.
	 */
	@Deprecated
	public static native void linkImage(byte[] sd, int imageReference, boolean freed);

	public static String getRawImagePath(String path) {
		return path + IMAGE_RAW_SUFFIX;
	}

	public static String getExternalImagePath(String path) {
		return path + "\0"; //$NON-NLS-1$
	}

	/**
	 * Throws an exception according the native return code.
	 *
	 * @param ret
	 *            the native return code
	 * @param path
	 *            the path of the image to load or null
	 * @return the native return code
	 */
	public static final int checkError(int ret, @Nullable String path) {

		int error;
		switch (ret) {
		case OUT_OF_MEMORY:
			// not enough room in images heap
			error = MicroUIException.IMAGE_OUT_OF_MEMORY;
			break;
		case INVALID_TYPE:
			// the native side cannot decode this kind of image
			error = MicroUIException.IMAGE_UNKNOWN_FORMAT;
			break;
		case INVALID_PATH:
			// cannot retrieve the image
			error = MicroUIException.RESOURCE_INVALID_PATH;
			break;
		case INVALID_FILE:
			// there was an error to open the png image
			error = MicroUIException.RESOURCE_INVALID_FILE;
			break;
		default:
			if (ret >= 0) {
				// no error
				return ret;
			}
			// unknown error
			error = ret;
			break;
		}

		throw new MicroUIException(error, path);
	}
}