/*
 * Java
 *
 * Copyright 2010-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import com.is2t.microbsp.microui.natives.NSystemDisplay;
import com.is2t.tools.GeometryTools;

import ej.bon.XMath;
import ej.microui.MicroUI;

/**
 * microUI-API
 */
public class GraphicsContext {

	/**
	 * microUI-API
	 */
	public static final int OPAQUE = 0xff;

	/**
	 * microUI-API
	 */
	public static final int TRANSPARENT = 0x0;

	private final SImageMetadata desc;

	/*
	 * These values are a copy of values available in {@link SImageMetadata}. There
	 * are used to inline the calls.
	 */
	private final int width;
	private final int height;

	/* default */ int translateX;
	// current translation ABSOLUTE LOCATION
	/* default */ int translateY;

	/*
	 * These values are a copy of values available in {@link SImageMetadata}. There
	 * are used to inline the calls. They MUST NOT be updated outside setClip
	 * methods.
	 */
	private int graphicalEngineClipX;
	private int graphicalEngineClipY;
	private int graphicalEngineClipWidth;
	private int graphicalEngineClipHeight;

	/**
	 * microUI-API
	 */
	@SuppressWarnings("unused")
	private GraphicsContext() {
		this(null);
	}

	/* default */ GraphicsContext(byte[] sd) {
		this(sd, MicroUI.isStarted());
	}

	/* default */ GraphicsContext(byte[] sd, boolean resetContext) {
		this.desc = new SImageMetadata(sd);
		this.width = this.desc.getWidth();
		this.height = this.desc.getHeight();
		if (resetContext) {
			resetContext();
		}
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getWidth() {
		return this.width;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getHeight() {
		return this.height;
	}

	/**
	 * microUI-API
	 *
	 * @param opacityPercent
	 *            microUI-API
	 * @return microUI-API
	 */
	public static int getAlpha(int opacityPercent) {
		return XMath.limit(opacityPercent * OPAQUE / 100, 0, OPAQUE);
	}

	/**
	 * microUI-API
	 */
	public void resetTranslate() {
		this.translateX = 0;
		this.translateY = 0;
	}

	/**
	 * microUI-API
	 */
	public void resetTranslate(int x, int y) {
		this.translateX = x;
		this.translateY = y;
	}

	/**
	 * microUI-API
	 *
	 * @param x
	 *            microUI-API
	 * @param y
	 *            microUI-API
	 */
	public void translate(int x, int y) {
		// change attributes. It is coherent to have translateXY < 0
		this.translateX += x;
		this.translateY += y;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getTranslateX() {
		return this.translateX;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getTranslateY() {
		return this.translateY;
	}

	/**
	 * microUI-API
	 *
	 * @param rgbColor
	 *            microUI-API
	 */
	public void setColor(int rgbColor) {
		getMetadata().setForegroundColor(rgbColor);
	}

	/**
	 * microUI-API
	 *
	 * @param rgbColor
	 *            microUI-API
	 */
	public void setBackgroundColor(int rgbColor) {
		NSystemDisplay.setBackgroundColor(this.getSNIContext(), rgbColor & ~SImageMetadata.ALPHA_MASK, true);
	}

	/**
	 * microUI-API
	 */
	public void removeBackgroundColor() {
		// change attribute in native world
		NSystemDisplay.setBackgroundColor(this.getSNIContext(), 0, false);
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getColor() {
		return getMetadata().getForegroundColor() & ~SImageMetadata.ALPHA_MASK;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getBackgroundColor() {
		return getMetadata().getBackgroundColor();
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public boolean hasBackgroundColor() {
		return getMetadata().hasBackgroundColor();
	}

	/**
	 * microUI-API
	 *
	 * @param x
	 *            microUI-API
	 * @param y
	 *            microUI-API
	 * @param width
	 *            microUI-API
	 * @param height
	 *            microUI-API
	 */
	public void setClip(int x, int y, int width, int height) {
		NSystemDisplay.setSubClip(getSNIContext(), x + this.translateX, y + this.translateY, width, height);
		updateJavaClip();
	}

	/**
	 * microUI-API
	 *
	 * @param x
	 *            microUI-API
	 * @param y
	 *            microUI-API
	 * @param width
	 *            microUI-API
	 * @param height
	 *            microUI-API
	 */
	public void resetClip(int x, int y, int width, int height) {
		NSystemDisplay.setClip(getSNIContext(), x + this.translateX, y + this.translateY, width, height);
		updateJavaClip();
	}

	/**
	 * microUI-API
	 */
	public void resetClip() {
		NSystemDisplay.resetClip(getSNIContext());
		updateJavaClip();
	}

	private void updateJavaClip() {
		this.graphicalEngineClipX = this.desc.getClipX1();
		this.graphicalEngineClipY = this.desc.getClipY1();
		this.graphicalEngineClipWidth = this.desc.getClipWidth();
		this.graphicalEngineClipHeight = this.desc.getClipHeight();
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getClipX() {
		// Separating the getClip operation into two methods returning
		// integers is more fast and memory efficient than one getClip() call
		// returning an object.
		return this.graphicalEngineClipX - this.translateX;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getClipY() {
		// Separating the getClip operation into two methods returning integers
		// is more performance and memory efficient than one getClip() call
		// returning an object.
		return this.graphicalEngineClipY - this.translateY;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getClipWidth() {
		return this.graphicalEngineClipWidth;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getClipHeight() {
		return this.graphicalEngineClipHeight;
	}

	/**
	 * microUI-API
	 *
	 * @param x
	 *            microUI-API
	 * @param y
	 *            microUI-API
	 * @return microUI-API
	 */
	public int readPixel(int x, int y) {
		x += this.translateX;
		y += this.translateY;
		if (x < 0 || y < 0 || x >= getMetadata().getWidth() || y >= getMetadata().getHeight()) {
			// test is performed here to throw the exception. It is better to perform this
			// test on native side but how to throw the exception on simulator ? task
			// M0092MEJAUI-1947
			throw new IllegalArgumentException();
		}
		return NSystemDisplay.readPixel(this.getSNIContext(), x, y);
	}

	/**
	 * microUI-API
	 *
	 * @param argbData
	 *            microUI-API
	 * @param offset
	 *            microUI-API
	 * @param scanlength
	 *            microUI-API
	 * @param x
	 *            microUI-API
	 * @param y
	 *            microUI-API
	 * @param width
	 *            microUI-API
	 * @param height
	 *            microUI-API
	 */
	public void readPixels(int[] argbData, int offset, int scanlength, int x, int y, int width, int height) {
		getARGB(this.desc, argbData, offset, scanlength, x, y, width, height);
	}

	/* default */ static void getARGB(SImageMetadata metadata, int[] argbData, int offset, int scanlength, int x, int y,
			int width, int height) {

		// check scan bounds
		GeometryTools.checkScanBounds(argbData.length, offset, scanlength, width, height);

		if (x < 0 || y < 0 || x + width - 1 > metadata.getWidth() || y + height - 1 > metadata.getHeight()) {
			throw new IllegalArgumentException();
		}
		NSystemDisplay.getARGB(metadata.getGraphicalEngineData(), argbData, offset, scanlength, x, y, width, height);
	}

	/**
	 * microUI-API
	 *
	 * @param width
	 *            microUI-API
	 */
	public void setEllipsisWidth(int width) {
		if (width <= 0) {
			throw new IllegalArgumentException();
		}
		getMetadata().setEllipsis((char) width);
	}

	/**
	 * microUI-API
	 */
	public void resetEllipsis() {
		getMetadata().setEllipsis((char) 0); // 0 means "no ellipsis"
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getEllipsisWidth() {
		return getMetadata().getEllipsis();
	}

	/**
	 * microUI-API
	 */
	public void resetDrawingParameters() {
		// spec: current color is black
		setColor(Colors.BLACK);
		removeBackgroundColor();
		resetEllipsis();
	}

	/**
	 * microUI-API
	 */
	public final void resetContext() {
		resetTranslate();
		resetClip();
	}

	/* default */ SImageMetadata getMetadata() {
		return this.desc;
	}

	/**
	 * microUI-API
	 *
	 * @return metadata
	 */
	public byte[] getSNIContext() {
		return this.desc.getGraphicalEngineData(); // XXX try to inline me
	}

}
