/*
 * Java
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import ej.bon.ByteArray;

/**
 * This class holds a byte array used by graphical engine to store the image
 * metadata.
 * <p>
 * Actually the graphical engine is using two structs: one for immutable image
 * and one for mutable image (which extends first struct and adding some
 * fields). This class should "represents" these both structs (not performed
 * yet).
 * <p>
 * On simulator there are some restrictions: the fields are stored in big endian
 * by graphical engine (== S3 endianness) but the API
 * {@link ByteArray#readInt(byte[], int)} reads an integer in platform
 * endianness (often little endian). So a patch is performed in
 * {@link #getInt(byte[], int)} (and others methods) for simulator.
 *
 */
public class SImageMetadata {

	/**
	 * Opacity mask
	 */
	public static final int ALPHA_MASK = 0xff000000;

	/**
	 * Upper part: immutable image
	 */
	private static final int OFFSET_I_OFFSCREEEN_ID = 0;
	private static final int OFFSET_C_WIDTH = OFFSET_I_OFFSCREEEN_ID + 4;
	private static final int OFFSET_C_HEIGHT = OFFSET_C_WIDTH + 2;
	private static final int OFFSET_C_MEMWH = OFFSET_C_HEIGHT + 2;
	private static final int OFFSET_B_FORMAT = OFFSET_C_MEMWH + 2;
	private static final int OFFSET_B_FLAGS = OFFSET_B_FORMAT + 1;
	private static final int SD_IMMUTABLE_SIZE = OFFSET_B_FLAGS + 1;

	/**
	 * Lower part: mutable image
	 */
	private static final int OFFSET_I_FOREGROUND_COLOR = SD_IMMUTABLE_SIZE + 0;
	private static final int OFFSET_I_BACKGROUND_COLOR = OFFSET_I_FOREGROUND_COLOR + 4;
	private static final int OFFSET_S_CLIPX1 = OFFSET_I_BACKGROUND_COLOR + 4;
	private static final int OFFSET_S_CLIPY1 = OFFSET_S_CLIPX1 + 2;
	private static final int OFFSET_S_CLIPX2 = OFFSET_S_CLIPY1 + 2;
	private static final int OFFSET_S_CLIPY2 = OFFSET_S_CLIPX2 + 2;
	private static final int OFFSET_C_ELLIPSIS = OFFSET_S_CLIPY2 + 2;
	private static final int SD_MUTABLE_SIZE = OFFSET_C_ELLIPSIS + 2;

	/**
	 * sync with SImageMetadata.icetea and sub struct
	 * <p>
	 * free: 5
	 */
	private static final int FLAG_MUTABLE = 1 << 0; // true when this targets a mutable image or display
	private static final int FLAG_FREED = 1 << 1; // true when the image data memory has to be freed (runtime allocator)
	private static final int FLAG_MAIN_SCREEN = 1 << 2; // true when this targets a main screen (not an image)
	private static final int FLAG_NO_HEADER = 1 << 3; // true when there is no header before the pixels array
	private static final int FLAG_CLOSED = 1 << 4; // true when the image has been closed
	private static final int FLAG_BACKCOLOR = 1 << 6; // true if there is a background color

	/**
	 * Icetea object: ScreenDescriptor
	 */
	private final byte[] graphicalEngineMetadata;

	/**
	 * Creates a new metadata with metadata previously filled by graphical engine.
	 */
	public SImageMetadata(byte[] graphicalEngineMetadata) {
		this.graphicalEngineMetadata = graphicalEngineMetadata;
	}

	/**
	 * Allocates a byte array to store the metadata of a mutable image or of an
	 * immutable image.
	 *
	 * @param mutable
	 *            true to allocate a byte array where storing the mutable image
	 *            metadata
	 * @return the allocated byte array
	 */
	public static byte[] allocate(boolean mutable) {
		return new byte[mutable ? SD_MUTABLE_SIZE : SD_IMMUTABLE_SIZE];
	}

	/**
	 * @return the graphical engine image metadata
	 */
	public byte[] getGraphicalEngineData() {
		return this.graphicalEngineMetadata;
	}

	/**
	 * Internal image reference in the graphical engine
	 *
	 * @return an integer which represents the image in the graphical engine.
	 */
	public int getImageReference() {
		return Tools.getInt(this.graphicalEngineMetadata, OFFSET_I_OFFSCREEEN_ID);
	}

	/**
	 * Tells if an image has to be freed in native world
	 *
	 * @return true if the image has to be freed (allocated in images heap).
	 *
	 * @deprecated Kept for backward compatibility (widget2 over UI3). Task
	 *             M0092MEJAUI-2065.
	 */
	@Deprecated
	public boolean isFreed() {
		return isFlagSet(FLAG_FREED);
	}

	public boolean hasBackgroundColor() {
		return isFlagSet(FLAG_BACKCOLOR);
	}

	public int getEllipsis() {
		return Tools.getChar(this.graphicalEngineMetadata, OFFSET_C_ELLIPSIS);
	}

	public int getWidth() {
		return Tools.getChar(this.graphicalEngineMetadata, OFFSET_C_WIDTH);
	}

	public int getHeight() {
		return Tools.getChar(this.graphicalEngineMetadata, OFFSET_C_HEIGHT);
	}

	public void setEllipsis(char width) {
		Tools.setChar(this.graphicalEngineMetadata, OFFSET_C_ELLIPSIS, width);
	}

	public void setForegroundColor(int color) {
		// change attribute in native world (force opaque pixel)
		Tools.setInt(this.graphicalEngineMetadata, OFFSET_I_FOREGROUND_COLOR, color | ALPHA_MASK);
	}

	public int getForegroundColor() {
		return Tools.getInt(this.graphicalEngineMetadata, OFFSET_I_FOREGROUND_COLOR);
	}

	public int getBackgroundColor() {
		return Tools.getInt(this.graphicalEngineMetadata, OFFSET_I_BACKGROUND_COLOR);
	}

	/**
	 * This method MUST ONLY BE called by {@link GraphicsContext#updateJavaClip()}
	 */
	public int getClipX1() {
		return Tools.getShort(this.graphicalEngineMetadata, OFFSET_S_CLIPX1);
	}

	/**
	 * This method MUST ONLY BE called by {@link GraphicsContext#updateJavaClip()}
	 */
	public int getClipY1() {
		return Tools.getShort(this.graphicalEngineMetadata, OFFSET_S_CLIPY1);
	}

	private int getClipX2() {
		return Tools.getShort(this.graphicalEngineMetadata, OFFSET_S_CLIPX2);
	}

	private int getClipY2() {
		return Tools.getShort(this.graphicalEngineMetadata, OFFSET_S_CLIPY2);
	}

	/**
	 * This method MUST ONLY BE called by {@link GraphicsContext#updateJavaClip()}
	 */
	public int getClipWidth() {
		return getClipX2() - getClipX1() + 1;
	}

	/**
	 * This method MUST ONLY BE called by {@link GraphicsContext#updateJavaClip()}
	 */
	public int getClipHeight() {
		return getClipY2() - getClipY1() + 1;
	}

	private boolean isFlagSet(int flag) {
		return (getGraphicalEngineData()[OFFSET_B_FLAGS] & flag) == flag;
	}
}
