/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector;

/**
 * A 2D transformation matrix.
 */
public class Matrix {

	/** The flat transformation matrix [m00 m01 m02 m10 m11 m12 m20 m21 m22]. */
	public final float[] values;

	/**
	 * Creates a new instance of transformation matrix. All the values are zero.
	 */
	public Matrix() {
		this.values = new float[9];
	}

	/**
	 * Sets this matrix to an identity matrix.
	 */
	public void identity() {
		this.values[0] = 1.0f;
		this.values[1] = 0.0f;
		this.values[2] = 0.0f;
		this.values[3] = 0.0f;
		this.values[4] = 1.0f;
		this.values[5] = 0.0f;
		this.values[6] = 0.0f;
		this.values[7] = 0.0f;
		this.values[8] = 1.0f;
	}

	/**
	 * Sets this matrix to an identity matrix with a translation of (x, y) vector.
	 *
	 * @param x
	 *            the translation on x axis
	 * @param y
	 *            the translation on y axis
	 */
	public void init(float x, float y) {
		this.values[0] = 1.0f;
		this.values[1] = 0.0f;
		this.values[2] = x;
		this.values[3] = 0.0f;
		this.values[4] = 1.0f;
		this.values[5] = y;
		this.values[6] = 0.0f;
		this.values[7] = 0.0f;
		this.values[8] = 1.0f;
	}

	/**
	 * Translates this matrix with the vector (x, y).
	 *
	 * @param x
	 *            the translation on x axis
	 * @param y
	 *            the translation on y axis
	 */
	public void translate(float x, float y) {
		this.values[2] = this.values[0] * x + this.values[1] * y + this.values[2];
		this.values[5] = this.values[3] * x + this.values[4] * y + this.values[5];
		this.values[8] = this.values[6] * x + this.values[7] * y + this.values[8];
	}

	/**
	 * Rotates this matrix with the given angle in degrees.
	 *
	 * @param angleDegrees
	 *            the rotation angle in degrees
	 */
	public void rotate(float angleDegrees) {
		final double angleRadians = Math.toRadians(angleDegrees);

		float cosAngle = (float) Math.cos(angleRadians);
		float sinAngle = (float) Math.sin(angleRadians);

		float tmp;

		tmp = cosAngle * this.values[0] + sinAngle * this.values[1];
		this.values[1] = cosAngle * this.values[1] - sinAngle * this.values[0];
		this.values[0] = tmp;

		tmp = cosAngle * this.values[3] + sinAngle * this.values[4];
		this.values[4] = cosAngle * this.values[4] - sinAngle * this.values[3];
		this.values[3] = tmp;

		tmp = cosAngle * this.values[6] + sinAngle * this.values[7];
		this.values[7] = cosAngle * this.values[7] - sinAngle * this.values[6];
		this.values[6] = tmp;
	}

	/**
	 * Scales this matrix.
	 *
	 * @param scaleX
	 *            the scale to apply on x axis
	 * @param scaleY
	 *            the scale to apply on y axis
	 */
	public void scale(float scaleX, float scaleY) {
		this.values[0] *= scaleX;
		this.values[1] *= scaleY;
		this.values[3] *= scaleX;
		this.values[4] *= scaleY;
		this.values[6] *= scaleX;
		this.values[7] *= scaleY;
	}
}
