/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector.font;

import ej.microui.display.GraphicsContext;

/**
 * A multiple linear gradient to apply to fonts.
 * <p>
 * Note: {@link LinearGradient} objects are not tread safe and may be modified by calling {@link VectorFontPainter}
 * drawing methods. Thus, the same gradient should not be used in several threads.
 * <p>
 * Each created instance must be closed, calling {@link #close()} method, to free memory.
 */
public class LinearGradient implements AutoCloseable {

	/* package */ final int handle;

	private int translateX;

	private int translateY;

	private boolean closed;

	/**
	 * Creates a new gradient instance.
	 *
	 * @param startX
	 *            the x coordinate of the start of the gradient
	 * @param startY
	 *            the y coordinate of the start of the gradient
	 * @param endX
	 *            the x coordinate of the end of the gradient
	 * @param endY
	 *            the y coordinate of the end of the gradient
	 * @param fractions
	 *            the colors distribution along the gradient from 0 (start) to 255 (end)
	 * @param colors
	 *            the colors in ARGB format
	 */
	public LinearGradient(int startX, int startY, int endX, int endY, int[] fractions, int[] colors) {
		this.handle = initNative(startX, startY, endX, endY, fractions, colors);
		this.translateX = 0;
		this.translateY = 0;
	}

	/**
	 * Sets the translation vector of this graphics context to the (x,y) vector.
	 *
	 * This method is used to update gradient according to the {@link GraphicsContext} translation vector.
	 *
	 * @param x
	 *            the x coordinate of the translation vector, in pixels.
	 * @param y
	 *            the y coordinate of the translation vector, in pixels.
	 * @see GraphicsContext#resetTranslate(int, int)
	 */
	protected void resetTranslate(int x, int y) {
		int shiftX = x - this.translateX;
		int shiftY = y - this.translateY;

		if (shiftX != 0 && shiftY != 0) {
			translateNative(this.handle, shiftX, shiftY);
			this.translateX = x;
			this.translateY = y;
		}

	}

	@Override
	public void close() {
		if (this.closed) {
			return;
		}

		this.closed = true;
		closeNative(this.handle);
	}

	private static native int initNative(int startX, int startY, int endX, int endY, int[] fractions, int[] colors);

	private static native void translateNative(int handle, int x, int y);

	private static native void closeNative(int handle);
}
